function s=qd(p1,p2,p3,p4)
%qd:  --  Construct quad-double (QD) Objects
%
%   S=qd(a) constructs the object S, of class qd from a.
%   If the input argument is a numerical scalar or matrix of type double, the 
%   result is a qd object having the given value for the p1 (high) field and 
%   p2,p3 and p4 (lower) fields set to zeros.
%
%   If input is a qd object, it is returned as it is.
%
%   S=qd(p1,p2,p3,p4) sets the highest field of the output to be equal to p1, and the lowest to p4.
%

%   Also see 
superiorto('dd','double','single','logical');
inferiorto('intval','gradient');
switch nargin
    case 0
        s.p1=[];s.p2=[];s.p3=[];s.p4=[];
        s=class(s,'qd');
        return
    case 1
        %just one argument
        if isa(p1,'qd') %already of the specified data type
            s=p1;
            return
        end
        if isempty(p1)
            s.p1=[];s.p2=[];s.p3=[];s.p4=[];
            s=class(s,'qd');
            return
        end
        if isnumeric(p1) | islogical(p1)
            %any numeric type is valid; convert it to double and set 0.0 for the low part
            s.p1=double(p1);
            p2=builtin('zeros',size(p1));
            %deal with NaN and Inf's
            q=find(isnan(p1(:)));if ~isempty(q),p2(q)=NaN;end
            q=find(isinf(p1(:)));if ~isempty(q),p2(q)=p1(q);end
            s.p2=p2;s.p3=p2;s.p4=p2;
            s=class(s,'qd');
            return
        else
            %input is not numeric!
            error(['Invalid data type for input: ' class(p1)])
        end
    case 2
        %assume to receive a dd-type
        if isempty(p2) & ~isempty(p1)
            error('Second argument should not be empty')
        end
        if isempty(p2) & isempty(p1)
            s.p1=[];s.p2=[];s.p3=[];s.p4=[];
            s=class(s,'qd');
            return
        end
        if isempty(p1)
            error('First argument should not be empty if two arguments are provided')
        end
        if ~all(size(p1)==size(p2))
            error('Both arguments should have the same size')
        end
        if isa(p1,'qd') | isa(p2,'qd') | isa(p1,'dd') | isa(p2,'dd')
            error('Inputs should not be of type dd/qd')
        end
        if isnumeric(p1) & isnumeric(p2)
            %Use Yozo's trick, relying in the renormalization routine already built-in in
            %the addition
            s=qd(p1)+p2;
            return
        else
            %input is not numeric!
            error(['Invalid data type for input qd(p1,p2): p1(' class(p1) ') and p2(' class(p2) ')'])
        end
    case 4
        %assume to receive only double-type
        if isempty(p2) & ~isempty(p1)
            error('Second argument should not be empty')
        end
        if isempty(p1) & isempty(p2) & isempty(p3) & isempty(p4)
            s.p1=[];s.p2=[];s.p3=[];s.p4=[];
            s=class(s,'qd');
            return
        end
        if isempty(p1) | isempty(p2) | isempty(p3) | isempty(p4)
            error('Arguments should not be empty if four arguments are provided')
        end
        S1=size(p1);S2=size(p2);S3=size(p3);S4=size(p4);
        if ~all(S1==S2 & S1==S3 & S1==S4)
            error('Four arguments should have the same size')
        end
        S1=class(p1);S2=class(p2);S3=class(p3);S4=class(p4);
        if isa(p1,'qd') | isa(p2,'qd') | isa(p3,'qd') | isa(p4,'qd') | isa(p1,'dd') | isa(p2,'dd') | isa(p3,'dd') | isa(p4,'dd')
            error('Inputs should not be of type dd/qd')
        end
        if isnumeric(p1) & isnumeric(p2) & isnumeric(p3) & isnumeric(p4)
            %Use Yozo's trick, relying in the renormalization routine already built-in in
            %the addition
            s=((qd(p1)+p2)+p3)+p4;
            return
        else
            %input is not numeric!
            error(['Invalid data type for input qd(p1,p2,p3,p4): p1(' class(p1) ') p2(' class(p2) ') p3(' class(p3) ') and p4(' class(p4) ')'])
        end
    otherwise
        %invalid number of inputs
        error(['Invalid number of inputs for qd'])        
end