% This toolbox implements a new class in matlab, already available in C and FORTRAN
% It adds more representable numbers roughly in the same interval as the double's, thus allowing
% less roundoff and cancellation in typical calculations. It could be used with already
% existing code with a few changes if any. 
% Version 0.0, Nov 2006
%
%%%%%%%%%%%%%%%%%%%%%%%%%%
%   What is quad-double?
%%%%%%%%%%%%%%%%%%%%%%%%%%
% The trick is to store any valid dd-number as an unevaluated sum of four doubles. We will denote
% as (p1,p2,p3,p4) each of the elements. The relationship between them is such that:
%     fl(p1+p2+p3+p4)=fl(p1); fl(p2+p3+p4)=fl(p2);fl(p3+p4)=fl(p3);
% so they can easily store the number 1+10^-100 and alike. This might have a relevant impact
% in dot product operations, like scalar product of vectors, as well as evaluation of the
% norm of a vector when the entries are wildly different in size.
% NOTE:
% Notice that the largest number which could be represented is nearly the same as the one of 
% double-type, while the minimum non-zero will be equal. So this solution might be useful if 
% you need more numbers in the present representable interval, but might not be so if you need 
% to expand the interval.
%
%%%%%%%%%%%%%%%%%%
%   How to use it?
%%%%%%%%%%%%%%%%%%
% Using the library should not be too difficult. For example, if you have a code like
%  function out=myf(a,b,c)
%  out=sqrt(a)+exp(b)-c;
%
%  and you plan to invoke it with a=1e10, b=1 and c=1e-10, you will obtain using
%  standard double representation the answer
%  out =
%
%    1.000027182818284e+005
%
%  If you want more precision, you can invoke the same routine with just a small change
%      out=myf(qd(a),qd(b),qd(c)) 
%  and produce as an output
% 	out (hi) =
%      
%         1.000027182818284e+005
% 	
% 	out (lo) =
%      
%        -2.750979816294869e-012
%
%  Another more illustrative example is to evaluate z=sqrt(117)-sqrt(116), which suffers from
%  catastrophic cancellation due to the substraction of two nearly similar numbers. 
%  A far better solution is to modify slightly the expression by multiplying and dividing by 
%  sqrt(117)+sqrt(116), producing z=(117-116)/(sqrt(117)+sqrt(116)) without cancellation.
%  This is not an universal solution, but serves the purpose of having at hand a good estimate 
%  of the true value.
%
%  Using double, the discrepancy to the "cancellation-less" result is 3.042e-012. If we instead use
%  z=sqrt(qd(117))-sqrt(qd(116)) and compare again the corresponding "cancellation-less" value, the 
%  relative error is 1.347e-028
%
%  Once you invoke the routine with inputs of type qd, all the intermediate calculations will
%  automatically be promoted to the new class. However, some caveats should be consider (see below)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  How to modify a source code if an error ocurs:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  There are some unavoidable problems in using operator overloading. For example:
%          qdpi=4*atan(qd(1)); %double-double version of pi
%          H=zeros(2,2);
%          H(1,1)=qdpi;
%  will render an error 
%          ??? Conversion to double from qd is not possible.
%  because H has not been defined as qd. So every ocurrence of zeros, ones, rand, eps, pi and even [] 
%  should be replaced with the qd-equivalent versions to avoid such problems qd_zeros, qd_ones, etc.
%  Another trick, which renders the code usable for either qd-type or standard double inputs, is 
%  to do the following:
%    H=zeros(2,2)+0*qdpi;
%  thus H inherits the data type of qdpi. Similar solution is valid for ones, rand, etc. but not for
%  []; in such case, you should change the code from 
%    a=[];
%  to 
%    a=[]*qdpi;
%  All this stuff is due to the lack of automatic promotion in the current version of Matlab; it forces
%  the rule that the LHS should stay the same data type as before the assignment, and this could not
%  be overriden at present.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  How to install the library:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Simply expand the contents of the zip file, and aqd the directory NewClasses to the matlab path.
%  It has been tested in Matlab 6.5 under Windows, and compiled with MSVC++ 6.0 (through the mex 
%  command) with default parameters. See generarmexqd.m for them.
%  If you need to re-compile it you should read before the qd/dd library documentation to consider
%  relevant compiler flags for your system
%
%%%%%%%%%%%%%%%%%%%%
%  Acknowledgements:
%%%%%%%%%%%%%%%%%%%%
%  The package has been written by Carlos Lopez-Vazquez, and is built on top of the qd/dd library 
%  as described by the paper:
%
%     Quad-Double Arithmetic: Algorithms, Implementation, and Application
%     Yozo Hida, Xiaoye S. Li and David H. Bailey, Technical Report LBNL-46996, October 2000
%
%  We have received assistance from Antonio Lopez, Marcel Leutenegger and Yozo Hida, all of
%  which is gratefully acknowledged.
%
%%%%%%%%%%%%%%%%%%%%%
%  Known limitations:
%%%%%%%%%%%%%%%%%%%%%
%  We have yet to write the routines to solve A\b, eig(A), svd(A), A^X, etc. for the general case.
%  Some of them require accomodate the case of qd-complex variables, presently outside our scope.
%  At most we provided home-made solutions for some of them (\, eig). Any help is welcomed. 
%  No attempt to consider sparse neither complex variables has been done.
%  Inputs having Inf and Nan entries have been considered in some routines, but not in all of them.
%  The sprintf command does not recognize the format itself; it always uses the double one, but at least
%  provides back a string useful for many purposes.
%
%%%%%%%%%%%%%%%%%%%%%%%
%  Possible extensions:
%%%%%%%%%%%%%%%%%%%%%%%
%  It is believed that the same template could accomodate the case of quad-double representation.
%  This implies that any number is stored as the unevaluated sum of four doubles.
%
%%%%%%%%%%%%%%%%%%%
%  Tests performed:
%%%%%%%%%%%%%%%%%%%  
%  We have created a test_qd.m in order to find gross and trivial errors in the implementation and
%  to systematically show the improvement due to the different data model.
%  Performance measurement was outside our focus for the moment, so it is open for future work.
%  Correctness tests has NOT been performed, and it is certainly a limitation.

% By Carlos Lopez-Vazquez, carlos.lopez@ieee.org